#include "cnasica.h"
#include "data.h"
#include "dataget.h"
#include "label.h"
#include "stack.h"
#include "parse.h"
#include "function.h"
#include "execute.h"

#define	COMMAND_EMPTY		0
#define	COMMAND_END			1
#define	COMMAND_GOTO		2
#define	COMMAND_GOSUB		3
#define	COMMAND_RETURN		4
#define	COMMAND_DECLARE		5
#define	COMMAND_PRINT		6
#define	COMMAND_INPUT		7
#define	COMMAND_IF			8
#define	COMMAND_REPEAT		9
#define COMMAND_UNTIL		10
#define	COMMAND_OPEN		11
#define	COMMAND_CLOSE		12
#define	COMMAND_READ		13
#define	COMMAND_WRITE		14
#define	COMMAND_SEEK		15
#define	COMMAND_TELL		16

#define	MODE_READ			0
#define	MODE_WRITE			1
#define	MODE_READWRITE		2

#define	ERROR_NONE			0
#define	ERROR_CANT_OPEN		1
#define	ERROR_NOT_OPEN		2
#define	ERROR_EOF			3

char *commandlist[] = {
	"END",
	"GOTO",
	"GOSUB",
	"RETURN",
	"DECLARE",
	"PRINT",
	"INPUT",
	"IF",
	"REPEAT",
	"UNTIL",
	"OPEN",
	"CLOSE",
	"READ",
	"WRITE",
	"SEEK",
	"TELL",
	""
};

void Error(SData *data, char *filename, char *s, SLabel *labellist, SVariable *variablelist) {
	int line=1,i;

	for(i=0;i<data->ptr;i++) {
		if(data->data[i]==0x0a) line++;
	}

	printf("Error in %s, line %d: %s\n",filename,line,s);

	DataFree(data);
	LabelFree(labellist);
	VariableFree(variablelist);

	exit(1);
}

int GetCommand(char *s, int *x) {
	int i=0;

	if(s[0]==0) return COMMAND_EMPTY;
	while(stricmp(commandlist[i],s)&&commandlist[i][0]!=0) i++;
	if(commandlist[i][0]!=0) {
		*x = i + 1;
		return 0;
	}

	return -1;
}

int ParseCompare(SVariable *variablelist, SData *data, int *result) {
	char ident[MAX_IDENT];
	char op[MAX_IDENT];
	int tmpint,tmpint2;
	double tmpdouble,tmpdouble2;
	char tmpstr[MAX_STRING];
	char tmpstr2[MAX_STRING];

	JumpSpace(data);
	if(data->data[data->ptr]!='(') return -1;
	data->ptr++;
	DataGetIdent(data,ident);
	JumpSpace(data);
	if(data->data[data->ptr]!=')') return -1;
	data->ptr++;
	JumpSpace(data);

	if(!stricmp(ident,"INTEGER")) {
		if(ParseInt(variablelist,data,&tmpint)==-1) return -1;

		JumpSpace(data);
		DataGetOperator(data,op);
		JumpSpace(data);

		if(ParseInt(variablelist,data,&tmpint2)==-1) return -1;

		if(!strcmp(op,"=")) {
			if(tmpint==tmpint2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<>")) {
			if(tmpint!=tmpint2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<")) {
			if(tmpint<tmpint2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,">")) {
			if(tmpint>tmpint2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<=")) {
			if(tmpint<=tmpint2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,">=")) {
			if(tmpint>=tmpint2) *result = 1; else *result = 0;
			return 0;
		} else return -1;
	} else
	if(!stricmp(ident,"FLOAT")) {
		if(ParseFloat(variablelist,data,&tmpdouble)==-1) return -1;

		JumpSpace(data);
		DataGetOperator(data,op);
		JumpSpace(data);

		if(ParseFloat(variablelist,data,&tmpdouble2)==-1) return -1;

		if(!strcmp(op,"=")) {
			if(tmpdouble==tmpdouble2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<>")) {
			if(tmpdouble!=tmpdouble2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<")) {
			if(tmpdouble<tmpdouble2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,">")) {
			if(tmpdouble>tmpdouble2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<=")) {
			if(tmpdouble<=tmpdouble2) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,">=")) {
			if(tmpdouble>=tmpdouble2) *result = 1; else *result = 0;
			return 0;
		} else return -1;
	} else
	if(!stricmp(ident,"STRING")) {
		if(ParseString(variablelist,data,tmpstr)==-1) return -1;

		JumpSpace(data);
		DataGetOperator(data,op);
		JumpSpace(data);

		if(ParseString(variablelist,data,tmpstr2)==-1) return -1;

		if(!strcmp(op,"=")) {
			if(!strcmp(tmpstr,tmpstr2)) *result = 1; else *result = 0;
			return 0;
		} else
		if(!strcmp(op,"<>")) {
			if(strcmp(tmpstr,tmpstr2)) *result = 1; else *result = 0;
			return 0;
		} else return -1;		
	}
	return -1;
}

int Execute(SData *data, char *filename) {
	char ident[MAX_IDENT];
	char ident2[MAX_IDENT];
	char tmpstr[MAX_STRING];
	int labels;
	int command;
	SLabel *labellist;
	SVariable *variablelist = NULL;
	SVariable *variable;
	UVariableValue value;
	int tmpint,tmpint2,tmpint3;
	double tmpdouble;
	FILE *file[MAX_FILES];

	for(tmpint=0;tmpint<MAX_FILES;tmpint++) file[tmpint] = NULL;
	value.integer = ERROR_NONE;
	VariableCreate(&variablelist, "ERROR", &value, VARIABLE_INTEGER);

	JumpSpace(data);
	DataGetIdent(data,ident);
	labellist = LabelLoad(data,&labels);
	data->ptr = 0;

	JumpSpace(data);
	while(data->data[data->ptr]!=0) {
		JumpSpace(data);
		while(is_line_break(data->data[data->ptr])) {
			JumpLineBreak(data);
			JumpSpace(data);
		}
		DataGetIdent(data,ident);
		JumpSpace(data);
		if(data->data[data->ptr]=='\'') {
			JumpToLineBreak(data);
			JumpLineBreak(data);
		} else
		if(GetCommand(ident,&command)==-1) {
			if(data->data[data->ptr]==':') {			
				data->ptr++;
			} else
			if(data->data[data->ptr]=='=') {
				data->ptr++;
				JumpSpace(data);
				if(VariableGet(variablelist,ident,&value)==-1) {
				}
				if((variable=VariablePointer(variablelist,ident)) == NULL) {
					Error(data,filename,"Variable not declared",labellist,variablelist);
				}
				switch(variable->type) {
					case VARIABLE_INTEGER:
						if(ParseInt(variablelist,data,&tmpint)==-1) {
							Error(data,filename,"Parse error",labellist,variablelist);
						}
						variable->value.integer = tmpint;
						break;
					case VARIABLE_FLOATING:
						if(ParseFloat(variablelist,data,&tmpdouble)==-1) {
							Error(data,filename,"Parse error",labellist,variablelist);
						}
						variable->value.floating = tmpdouble;
						break;
					case VARIABLE_STRING:
						if(ParseString(variablelist,data,variable->value.string)==-1) {
							Error(data,filename,"Parse error",labellist,variablelist);
						}
						break;
					default:						
						Error(data,filename,"Variable type not supported",labellist,variablelist);
				}
			} else {
				Error(data,filename,"Unknown command",labellist,variablelist);
			}
		} else {
		switch(command) {
			case COMMAND_EMPTY:
				Error(data,filename,"Syntax error",labellist,variablelist);
				break;
			case COMMAND_END:
				goto Quit;
			case COMMAND_GOTO:
				DataGetIdent(data,ident);
				if(LabelFind(labellist,labels,ident,&tmpint)==-1) {
					Error(data,filename,"Unknown label, GOTO failed",labellist,variablelist);
				}
				data->ptr = tmpint;
				break;
			case COMMAND_GOSUB:
				DataGetIdent(data,ident);
				if(LabelFind(labellist,labels,ident,&tmpint)==-1) {
					Error(data,filename,"Unknown label, GOSUB failed",labellist,variablelist);
				}
				if(StackPush(data->ptr,STACK_GOSUB)==-1) {
					Error(data,filename,"Stack overflow",labellist,variablelist);
				}
				data->ptr = tmpint;
				break;
			case COMMAND_RETURN:
				if(StackPop(&tmpint,&tmpint2)==-1) {
					Error(data,filename,"RETURN without GOSUB",labellist,variablelist);
				}
				if(tmpint2!=STACK_GOSUB) {
					Error(data,filename,"RETURN value not of GOSUB type",labellist,variablelist);
				}
				data->ptr = tmpint;
				break;
			case COMMAND_DECLARE:
				do {
					if(data->data[data->ptr]==',') {
						data->ptr++;
						JumpSpace(data);
					}
					DataGetIdent(data,ident);
					if(CheckChar(data,':')==-1) {
						Error(data,filename,"Expected : in DECLARE command",labellist,variablelist);
					}
					DataGetIdent(data,ident2);
					JumpSpace(data);
					if(!stricmp(ident2,"INTEGER")) {
						value.integer = 0;
						VariableCreate(&variablelist,ident,&value,VARIABLE_INTEGER);
					} else
					if(!stricmp(ident2,"FLOAT")) {
						value.floating = 0;
						VariableCreate(&variablelist,ident,&value,VARIABLE_FLOATING);
					} else
					if(!stricmp(ident2,"STRING")) {
						VariableCreate(&variablelist,ident,&value,VARIABLE_STRING);
					} else {
						Error(data,filename,"Invalid data type",labellist,variablelist);
					}
				} while(data->data[data->ptr]==',');
				break;
			case COMMAND_PRINT:
				if(ParseString(variablelist,data,tmpstr)==-1) {
					Error(data,filename,"Parse error",labellist,variablelist);
				}
				printf("%s",tmpstr);
				JumpSpace(data);
				if(data->data[data->ptr]!=';') {
					printf("\n");
				} else {
					data->ptr++;
				}
				break;
			case COMMAND_INPUT:
				DataGetIdent(data,ident);
				if((variable=VariablePointer(variablelist,ident)) == NULL) {
					Error(data,filename,"Can not find the variable to input",labellist,variablelist);
				}
				switch(variable->type) {
					case VARIABLE_INTEGER:
						mygets(tmpstr);
						if(StrToInt(tmpstr,&variable->value.integer)==-1) {
							Error(data,filename,"Invalid number",labellist,variablelist);
						}
						break;
					case VARIABLE_FLOATING:
						mygets(tmpstr);
						if(StrToFloat(tmpstr,&variable->value.floating)==-1) {
							Error(data,filename,"Invalid number",labellist,variablelist);
						}
						break;
					case VARIABLE_STRING:
						mygets(variable->value.string);
						break;
				};
				break;
			case COMMAND_IF:
				if(ParseCompare(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Syntax error in compare",labellist,variablelist);
				}
				JumpSpace(data);
				DataGetIdent(data,ident);
				if(stricmp(ident,"THEN")) {
					Error(data,filename,"Expected THEN",labellist,variablelist);
				}
				JumpSpace(data);
				if(!tmpint) {
					JumpToLineBreak(data);
					JumpLineBreak(data);
				}
				break;
			case COMMAND_REPEAT:
				if(StackPush(data->ptr,STACK_REPEAT)==-1) {
					Error(data,filename,"Stack overflow",labellist,variablelist);
				}
				break;
			case COMMAND_UNTIL:
				if(StackPop(&tmpint,&tmpint2)==-1) {
					Error(data,filename,"UNTIL command filed, stack is empty",labellist,variablelist);
				}
				if(tmpint2!=STACK_REPEAT) {
					Error(data,filename,"Return pointer is not STACK_REPEAT",labellist,variablelist);
				}
				if(ParseCompare(variablelist,data,&tmpint2)==-1) {
					Error(data,filename,"Syntax error in compare",labellist,variablelist);
				}
				if(!tmpint2) {
					data->ptr = tmpint;
					if(StackPush(tmpint,STACK_REPEAT)==-1) {
						Error(data,filename,"Stack overflow",labellist,variablelist);
					}
				}
				break;
			case COMMAND_OPEN:
				if(ParseInt(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Can not read file nr for OPEN",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in OPEN, expected comma",labellist,variablelist);
				}
				if(ParseString(variablelist,data,tmpstr)==-1) {
					Error(data,filename,"Can not read file nr for OPEN",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in OPEN, expected comma",labellist,variablelist);
				}
				if(ParseInt(variablelist,data,&tmpint2)==-1) {
					Error(data,filename,"Can not mode nr for OPEN",labellist,variablelist);
				}
				if(tmpint>=MAX_FILES) {
					Error(data,filename,"Invalild file nr for OPEN",labellist,variablelist);
				}
				switch(tmpint2) {
					case MODE_READ:
						file[tmpint] = fopen(tmpstr,"rb");
						break;
					case MODE_WRITE:
						file[tmpint] = fopen(tmpstr,"wb");
						break;
					case MODE_READWRITE:
						file[tmpint] = fopen(tmpstr,"rb+");
						break;
					default:
						Error(data,filename,"Invalild file mode",labellist,variablelist);
						break;
				}
				if(file[tmpint]==NULL) {
					value.integer = ERROR_CANT_OPEN;
					if(VariableSet(variablelist,"ERROR",&value)==-1) {
						Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
					}
				}
				break;
			case COMMAND_CLOSE:
				if(ParseInt(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Can not read file nr for CLOSE",labellist,variablelist);
				}
				if(file[tmpint]!=NULL) {
					fclose(file[tmpint]);
				}
				break;
			case COMMAND_READ:
				if(ParseInt(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Can not read file nr for READ",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in READ, expected comma",labellist,variablelist);
				}
				DataGetIdent(data,ident);
				if((variable=VariablePointer(variablelist,ident)) == NULL) {
					Error(data,filename,"Can not find the variable to read from file",labellist,variablelist);
				}
				if(tmpint>=MAX_FILES) {
					Error(data,filename,"Invalild file nr for READ",labellist,variablelist);
				}
				if(file[tmpint]==NULL) {
					value.integer = ERROR_NOT_OPEN;
					if(VariableSet(variablelist,"ERROR",&value)==-1) {
						Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
					}
				} else {
				switch(variable->type) {
					case VARIABLE_INTEGER:
						if(fread(&variable->value.integer,1,sizeof(int),file[tmpint])!=sizeof(int)) {
							value.integer = ERROR_EOF;
							if(VariableSet(variablelist,"ERROR",&value)==-1) {
								Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
							}
						}
						break;
					case VARIABLE_FLOATING:
						if(fread(&variable->value.integer,1,sizeof(double),file[tmpint])!=sizeof(double)) {
							value.integer = ERROR_EOF;
							if(VariableSet(variablelist,"ERROR",&value)==-1) {
								Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
							}
						}
					case VARIABLE_STRING:
						fgets(variable->value.string,MAX_STRING,file[tmpint]);
						if(variable->value.string[strlen(variable->value.string)-1]=='\n')
							variable->value.string[strlen(variable->value.string)-1]=0;
						break;
				};
				}
				break;
			case COMMAND_WRITE:
				if(CheckChar(data,'(')==-1) {
					Error(data,filename,"Syntax error in WRITE, expected (",labellist,variablelist);
				}
				DataGetIdent(data,ident);
				if(CheckChar(data,')')==-1) {
					Error(data,filename,"Syntax error in WRITE, expected )",labellist,variablelist);
				}
				if(ParseInt(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Can not read file nr for WRITE",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in WRITE, expected comma",labellist,variablelist);
				}
				if(tmpint>=MAX_FILES) {
					Error(data,filename,"Invalild file nr for WRITE",labellist,variablelist);
				}
				if(file[tmpint]==NULL) {
					value.integer = ERROR_NOT_OPEN;
					if(VariableSet(variablelist,"ERROR",&value)==-1) {
						Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
					}
				} else {
				if(!stricmp(ident,"INTEGER")) {
					if(ParseInt(variablelist,data,&tmpint2)==-1) {
						Error(data,filename,"Parse error in WRITE",labellist,variablelist);
					}
					fwrite(&tmpint2,1,sizeof(int),file[tmpint]);
				} else
				if(!stricmp(ident,"FLOAT")) {
					if(ParseFloat(variablelist,data,&tmpdouble)==-1) {
						Error(data,filename,"Parse error in WRITE",labellist,variablelist);
					}
					fwrite(&tmpdouble,1,sizeof(double),file[tmpint]);
				} else
				if(!stricmp(ident,"STRING")) {
					if(ParseString(variablelist,data,tmpstr)==-1) {
						Error(data,filename,"Parse error in WRITE",labellist,variablelist);
					}
					fwrite(&tmpstr,1,strlen(tmpstr),file[tmpint]);
					if(CheckChar(data,';')==-1) {
						fprintf(file[tmpint],"\n");
					}
				} else {
					Error(data,filename,"Wrong type in WRITE",labellist,variablelist);
				}
				}
				break;
			case COMMAND_SEEK:
				if(ParseInt(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Can not read file nr for SEEK",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in SEEK, expected comma",labellist,variablelist);
				}
				if(ParseInt(variablelist,data,&tmpint2)==-1) {
					Error(data,filename,"Can not read offset for SEEK",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in SEEK, expected comma",labellist,variablelist);
				}
				if(ParseInt(variablelist,data,&tmpint3)==-1) {
					Error(data,filename,"Can not read seek type for SEEK",labellist,variablelist);
				}
				if(tmpint>=MAX_FILES) {
					Error(data,filename,"Invalild file nr for WRITE",labellist,variablelist);
				}
				if(file[tmpint]==NULL) {
					value.integer = ERROR_NOT_OPEN;
					if(VariableSet(variablelist,"ERROR",&value)==-1) {
						Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
					}
				} else {
				switch(tmpint3) {
					case 0:
						fseek(file[tmpint],tmpint2,SEEK_SET);
						break;
					case 1:
						fseek(file[tmpint],tmpint2,SEEK_CUR);
						break;
					case 2:
						fseek(file[tmpint],tmpint2,SEEK_END);
						break;
					default:
						Error(data,filename,"Invalid seek type for SEEK",labellist,variablelist);
						break;
				};
				}
				break;
			case COMMAND_TELL:
				if(ParseInt(variablelist,data,&tmpint)==-1) {
					Error(data,filename,"Can not read file nr for TELL",labellist,variablelist);
				}
				if(CheckChar(data,',')==-1) {
					Error(data,filename,"Syntax error in TELL, expected comma",labellist,variablelist);
				}
				DataGetIdent(data,ident);
				if((variable=VariablePointer(variablelist,ident)) == NULL) {
					Error(data,filename,"Can not find the variable to put file position in",labellist,variablelist);
				}
				if(tmpint>=MAX_FILES) {
					Error(data,filename,"Invalild file nr for TELL",labellist,variablelist);
				}
				if(file[tmpint]==NULL) {
					value.integer = ERROR_NOT_OPEN;
					if(VariableSet(variablelist,"ERROR",&value)==-1) {
						Error(data,filename,"Can not set ERROR variable",labellist,variablelist);
					}
				} else {
					if(variable->type!=VARIABLE_INTEGER) {
						Error(data,filename,"Invalid variable type (must be INTEGER)",labellist,variablelist);
					}
					variable->value.integer = ftell(file[tmpint]);
				}
				break;
			default:
				Error(data,filename,"Commandlist entry not supported (internal error)",labellist,variablelist);
				break;
		};
		}
		JumpSpace(data);
		while(is_line_break(data->data[data->ptr])) {
			JumpLineBreak(data);
			JumpSpace(data);
		}
	}

Quit:
	LabelFree(labellist);
	DataFree(data);
	VariableFree(variablelist);
	return 0;
}

int ExecuteFile(char *filename) {
	SData data;

	if(DataLoadFile(&data,filename)==-1) return -1;
	if(Execute(&data,filename)==-1) return -1;
	return 0;
}
