#include "cnasica.h"
#include "function.h"
#include "data.h"
#include "dataget.h"
#include "parse.h"

#define	TRIG_SIN	0
#define	TRIG_COS	1
#define	TRIG_TAN	2
#define	TRIG_ASIN	3
#define	TRIG_ACOS	4
#define	TRIG_ATAN	5

#define	LOG_LOG		0
#define	LOG_LN		1
#define	LOG_EXP		2

int CheckChar(SData *data, char c) {
	JumpSpace(data);
	if(data->data[data->ptr]!=c) {
		return -1;
	}
	data->ptr++;
	JumpSpace(data);

	return 0;
}

int StrGetFloat(SVariable *variablelist, char *s, double *x) {
	SVariable *param;

	if(is_float_dec(s)) {
		if(StrToFloat(s,x)==-1) return -1;
		return 0;
	}
	if((param=VariablePointer(variablelist,s)) == NULL) return -1;
	if(VariableGetFloat(param,x)==-1) return -1;

	return 0;
}

int StrGetInt(SVariable *variablelist, char *s, int *x) {
	SVariable *param;

	if(is_dec(s)) {
		if(StrToInt(s,x)==-1) return -1;
		return 0;
	}
	if((param=VariablePointer(variablelist,s)) == NULL) return -1;
	if(VariableGetInt(param,x)==-1) return -1;

	return 0;
}

int FunctionTrig(SData *data, SVariable *variablelist, SVariable *variable, int type) {
	double x;

	if(ParseFloat(variablelist,data,&x)==-1) return -1;

	switch(type) {
		case TRIG_SIN:
			variable->value.floating = (double) sin(x);
			break;
		case TRIG_COS:
			variable->value.floating = (double) cos(x);
			break;
		case TRIG_TAN:
			variable->value.floating = (double) tan(x);
			break;
		case TRIG_ASIN:
			variable->value.floating = (double) asin(x);
			break;
		case TRIG_ACOS:
			variable->value.floating = (double) acos(x);
			break;
		case TRIG_ATAN:
			variable->value.floating = (double) atan(x);
			break;
		default:
			return -1;
	}

	variable->type = VARIABLE_FLOATING;

	return 0;
}

int FunctionLog(SData *data, SVariable *variablelist, SVariable *variable, int type) {
	double x;

	if(ParseFloat(variablelist,data,&x)==-1) return -1;

	switch(type) {
		case LOG_LOG:
			variable->value.floating = (double) log10(x);
			break;
		case LOG_LN:
			variable->value.floating = (double) log(x);
			break;
		case LOG_EXP:
			variable->value.floating = (double) exp(x);
			break;
		default:
			return -1;
	}

	variable->type = VARIABLE_FLOATING;

	return 0;
}

int FunctionSqr(SData *data, SVariable *variablelist, SVariable *variable) {
	double x;

	if(ParseFloat(variablelist,data,&x)==-1) return -1;

	variable->value.floating = (double) x*x;
	variable->type = VARIABLE_FLOATING;

	return 0;
}

int FunctionSqrt(SData *data, SVariable *variablelist, SVariable *variable) {
	double x;

	if(ParseFloat(variablelist,data,&x)==-1) return -1;

	variable->value.floating = (double) sqrt(x);
	variable->type = VARIABLE_FLOATING;

	return 0;
}

int FunctionHex(SData *data, SVariable *variablelist, SVariable *variable) {
	int x;

	if(ParseInt(variablelist,data,&x)==-1) return -1;

	sprintf(variable->value.string,"%08x",x);
	variable->type = VARIABLE_STRING;

	return 0;
}

int FunctionLen(SData *data, SVariable *variablelist, SVariable *variable) {
	char s[MAX_STRING];

	if(ParseString(variablelist,data,s)==-1) return -1;

	variable->value.integer = strlen(s);
	variable->type = VARIABLE_INTEGER;

	return 0;
}

int FunctionAsc(SData *data, SVariable *variablelist, SVariable *variable) {
	char s[MAX_STRING];

	if(ParseString(variablelist,data,s)==-1) return -1;

	variable->value.integer = (int) s[0];
	variable->type = VARIABLE_INTEGER;

	return 0;
}

int FunctionMid(SData *data, SVariable *variablelist, SVariable *variable) {
	char s[MAX_STRING];
	int from,to,i;

	if(ParseInt(variablelist,data,&from)==-1) return -1;
	if(CheckChar(data,',')==-1) return -1;
	if(ParseInt(variablelist,data,&to)==-1) return -1;
	if(CheckChar(data,',')==-1) return -1;
	if(ParseString(variablelist,data,s)==-1) return -1;

	for(i=from;i<from+to;i++) variable->value.string[i-from] = s[i];
	variable->value.string[to]=0;
	variable->type = VARIABLE_STRING;

	return 0;
}

int FunctionPow(SData *data, SVariable *variablelist, SVariable *variable) {
	double x,y;

	if(ParseFloat(variablelist,data,&x)==-1) return -1;
	if(CheckChar(data,',')==-1) return -1;
	if(ParseFloat(variablelist,data,&y)==-1) return -1;

	variable->value.floating = (double) pow(x,y);
	variable->type = VARIABLE_FLOATING;

	return 0;
}

int FunctionExecute(SData *data, SVariable *variablelist, char *ident, SVariable *variable) {
	if(!stricmp(ident,"HEX")) {
		if(FunctionHex(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"LEN")) {
		if(FunctionLen(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"ASC")) {
		if(FunctionAsc(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"MID")) {
		if(FunctionMid(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"SQR")) {
		if(FunctionSqr(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"SQRT")) {
		if(FunctionSqrt(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"POW")) {
		if(FunctionPow(data,variablelist,variable)==-1) return -1;
	} else
	if(!stricmp(ident,"SIN")) {
		if(FunctionTrig(data,variablelist,variable,TRIG_SIN)==-1) return -1;
	} else
	if(!stricmp(ident,"COS")) {
		if(FunctionTrig(data,variablelist,variable,TRIG_COS)==-1) return -1;
	} else
	if(!stricmp(ident,"TAN")) {
		if(FunctionTrig(data,variablelist,variable,TRIG_TAN)==-1) return -1;
	} else
	if(!stricmp(ident,"ASIN")) {
		if(FunctionTrig(data,variablelist,variable,TRIG_ASIN)==-1) return -1;
	} else
	if(!stricmp(ident,"ACOS")) {
		if(FunctionTrig(data,variablelist,variable,TRIG_ACOS)==-1) return -1;
	} else
	if(!stricmp(ident,"ATAN")) {
		if(FunctionTrig(data,variablelist,variable,TRIG_ATAN)==-1) return -1;
	} else
	if(!stricmp(ident,"LOG")) {
		if(FunctionLog(data,variablelist,variable,LOG_LOG)==-1) return -1;
	} else
	if(!stricmp(ident,"LN")) {
		if(FunctionLog(data,variablelist,variable,LOG_LN)==-1) return -1;
	} else
	if(!stricmp(ident,"EXP")) {
		if(FunctionLog(data,variablelist,variable,LOG_EXP)==-1) return -1;
	} else {
		return -1;
	}
	JumpSpace(data);

	if(data->data[data->ptr]!=')') {
		return -1;
	}
	data->ptr++;

	return 0;
}
