#include "cnasica.h"
#include "parse.h"
#include "data.h"
#include "dataget.h"
#include "function.h"
#include "variable.h"

int is_float_dec(char *s) {
	int i,j,dot=0;

	j = strlen(s);
	for(i=0;i<j;i++) {
		if(!is_digit(s[i])&&s[i]!='.') {
			return 0;
		}
		if(s[i]=='.') dot++;
	}
	if(dot<=1) return 1;

	return 0;
}

int is_dec(char *s) {
	int i,j;

	j = strlen(s);
	for(i=0;i<j;i++) {
		if(!is_digit(s[i])) {
			return 0;
		}
	}

	return 1;
}

int StrToInt(char *s, int *result) {
	if(is_dec(s)) {
		*result = atoi(s);
		return 0;
	}
	return -1;
}

int StrToFloat(char *s, double *result) {
	if(is_float_dec(s)) {
		*result = atof(s);
		return 0;
	}
	return -1;
}

int ParseString(SVariable *variablelist, SData *data, char *result) {
	char tmpstr[MAX_STRING];
	char ident[MAX_IDENT];
	int i=0;
	char op;
	SVariable *variable;
	SVariable tmpvariable;

	result[0] = 0;

	do {
		JumpSpace(data);
		if(data->data[data->ptr]=='\"') {
			data->ptr++;
			while(data->data[data->ptr]!='\"') result[i++] = data->data[data->ptr++];
			data->ptr++;
			result[i]=0;
		} else {
			if(DataGetIdent(data,ident)==-1) return -1;
			JumpSpace(data);
			if(data->data[data->ptr]=='(') {
				data->ptr++;
				if(FunctionExecute(data,variablelist,ident,&tmpvariable)==-1) {
					return -1;
				}
				if(VariableGetString(&tmpvariable,tmpstr)==-1) return -1;
				strcat(result,tmpstr);
				i+=strlen(tmpstr);
			} else {
				if((variable=VariablePointer(variablelist,ident)) == NULL) return -1;
				if(VariableGetString(variable,tmpstr)==-1) return -1;
				strcat(result,tmpstr);
				i+=strlen(tmpstr);
			}
		}
		JumpSpace(data);
		op = data->data[data->ptr];
		if(op=='+') data->ptr++;
	} while(op=='+');

	return 0;
}

/* Parsing of integer numbers */

int ParseInt(SVariable *variablelist, SData *data, int *result) {
	if(ParseIntLevel1(variablelist,data,result)==-1) return -1;
	return 0;
}

int ParseIntLevel1(SVariable *variablelist, SData *data, int *result) {
	int hold;
	char op;

	JumpSpace(data);
	if(ParseIntLevel2(variablelist,data,result)==-1) return -1;
	JumpSpace(data);

	op = data->data[data->ptr];
	while(op=='+' || op=='-') {
		data->ptr++;
		if(ParseIntLevel2(variablelist,data,&hold)==-1) return -1;
		switch(op) {
			case '+':
				*result += hold;
				break;
			case '-':
				*result -= hold;
				break;
		};
		JumpSpace(data);
		op = data->data[data->ptr];
	}
	return 0;
}

int ParseIntLevel2(SVariable *variablelist, SData *data, int *result) {
	int hold;
	char op;

	JumpSpace(data);
	if(ParseIntLevel3(variablelist,data,result)==-1) return -1;
	JumpSpace(data);

	op = data->data[data->ptr];
	while(op=='*' || op=='/' || op=='%') {
		data->ptr++;
		if(ParseIntLevel3(variablelist,data,&hold)==-1) return -1;
		switch(op) {
			case '*':
				*result *= hold;
				break;
			case '/':
				*result /= hold;
				break;
			case '%':
				*result %= hold;
				break;
		};
		JumpSpace(data);
		op = data->data[data->ptr];
	}
	return 0;
}

int ParseIntLevel3(SVariable *variablelist, SData *data, int *result) {
	JumpSpace(data);
	if(data->data[data->ptr]=='(') {
		data->ptr++;
		if(ParseIntLevel1(variablelist,data,result)==-1) return -1;
		JumpSpace(data);
		if(data->data[data->ptr]!=')') {
			return -1;
		}
		data->ptr++;
	} else {
		if(ParseIntPrimitive(variablelist,data,result)==-1) return -1;
	}
	return 0;
}

int ParseIntPrimitive(SVariable *variablelist, SData *data, int *result) {
	char ident[MAX_IDENT];
	SVariable *variable;
	double tmpdouble;
	int x;

	JumpSpace(data);
	DataGetIdent(data,ident);
	JumpSpace(data);
	if(is_digit(ident[0])) {
		if(StrToInt(ident,result)==-1) {
			if(StrToFloat(ident,&tmpdouble)==-1) return -1;
			*result = (int) tmpdouble;
		}
	} else {
		if(data->data[data->ptr]=='(') {
			variable = (SVariable*) malloc(sizeof(SVariable));
			data->ptr++;
			if(FunctionExecute(data,variablelist,ident,variable)==-1) return -1;
			JumpSpace(data);

			switch(variable->type) {
				case VARIABLE_INTEGER:
					x = variable->value.integer;
					break;
				case VARIABLE_STRING:
					if(StrToInt(variable->value.string,&x)==-1) {
						free(variable->value.string);
						return -1;
					}
					free(variable->value.string);
					break;
				case VARIABLE_FLOATING:
					x = (int) variable->value.floating;
					break;
			};
			*result = x;
			free(variable);

			return 0;
		}
		if((variable=VariablePointer(variablelist,ident)) == NULL) {
			return -1;
		}
		switch(variable->type) {
			case VARIABLE_INTEGER:
				x = variable->value.integer;
				break;
			case VARIABLE_STRING:
				if(StrToInt(variable->value.string,&x)==-1) return -1;
				break;
			case VARIABLE_FLOATING:
				x = (int) variable->value.floating;
				break;
		};
		*result = x;
	}
	return 0;
}

/* Parsing of floating-point numbers */

int ParseFloat(SVariable *variablelist, SData *data, double *result) {
	if(ParseFloatLevel1(variablelist,data,result)==-1) return -1;
	return 0;
}

int ParseFloatLevel1(SVariable *variablelist, SData *data, double *result) {
	double hold;
	char op;

	JumpSpace(data);
	if(ParseFloatLevel2(variablelist,data,result)==-1) return -1;
	JumpSpace(data);

	op = data->data[data->ptr];
	while(op=='+' || op=='-') {
		data->ptr++;
		if(ParseFloatLevel2(variablelist,data,&hold)==-1) return -1;
		switch(op) {
			case '+':
				*result += hold;
				break;
			case '-':
				*result -= hold;
				break;
		};
		JumpSpace(data);
		op = data->data[data->ptr];
	}
	return 0;
}

int ParseFloatLevel2(SVariable *variablelist, SData *data, double *result) {
	double hold;
	char op;

	JumpSpace(data);
	if(ParseFloatLevel3(variablelist,data,result)==-1) return -1;
	JumpSpace(data);

	op = data->data[data->ptr];
	while(op=='*' || op=='/' || op=='%') {
		data->ptr++;
		if(ParseFloatLevel3(variablelist,data,&hold)==-1) return -1;
		switch(op) {
			case '*':
				*result *= hold;
				break;
			case '/':
				*result /= hold;
				break;
			case '%':
				return -1;
		};
		JumpSpace(data);
		op = data->data[data->ptr];
	}
	return 0;
}

int ParseFloatLevel3(SVariable *variablelist, SData *data, double *result) {
	JumpSpace(data);
	if(data->data[data->ptr]=='(') {
		data->ptr++;
		if(ParseFloatLevel1(variablelist,data,result)==-1) return -1;
		JumpSpace(data);
		if(data->data[data->ptr]!=')') {
			return -1;
		}
		data->ptr++;
	} else {
		if(ParseFloatPrimitive(variablelist,data,result)==-1) return -1;
	}
	return 0;
}

int ParseFloatPrimitive(SVariable *variablelist, SData *data, double *result) {
	char ident[MAX_IDENT];
	SVariable *variable;
	double x;

	JumpSpace(data);
	DataGetIdent(data,ident);
	JumpSpace(data);
	if(is_digit(ident[0])) {
		if(StrToFloat(ident,result)==-1) return -1;
	} else {
		if(data->data[data->ptr]=='(') {
			variable = (SVariable*) malloc(sizeof(SVariable));
			data->ptr++;
			if(FunctionExecute(data,variablelist,ident,variable)==-1) return -1;
			JumpSpace(data);

			switch(variable->type) {
				case VARIABLE_INTEGER:
					x = (double) variable->value.integer;
					break;
				case VARIABLE_STRING:
					if(StrToFloat(variable->value.string,&x)==-1) {
						free(variable->value.string);
						return -1;
					}
					free(variable->value.string);
					break;
				case VARIABLE_FLOATING:
					x = variable->value.floating;
					break;
			};
			*result = x;
			free(variable);

			return 0;
		}
		if((variable=VariablePointer(variablelist,ident)) == NULL) {
			return -1;
		}
		switch(variable->type) {
			case VARIABLE_INTEGER:
				x = (double) variable->value.integer;
				break;
			case VARIABLE_STRING:
				if(StrToFloat(variable->value.string,&x)==-1) return -1;
				break;
			case VARIABLE_FLOATING:
				x = variable->value.floating;
				break;
		};
		*result = x;
	}
	return 0;
}
