#include "cnasica.h"
#include "variable.h"
#include "parse.h"

int VariableGetFloat(SVariable *variable, double *x) {
	switch(variable->type) {
		case VARIABLE_INTEGER:
			*x = (double) variable->value.integer;
			return 0;
		case VARIABLE_FLOATING:
			*x = variable->value.floating;
			return 0;
		case VARIABLE_STRING:
			if(StrToFloat(variable->value.string,x)==-1) return -1;
			*x = variable->value.floating;
			return 0;
	};

	return -1;
}

int VariableGetInt(SVariable *variable, int *x) {
	switch(variable->type) {
		case VARIABLE_INTEGER:
			*x = variable->value.integer;
			return 0;
		case VARIABLE_FLOATING:
			*x = (int) variable->value.floating;
			return 0;
		case VARIABLE_STRING:
			if(StrToInt(variable->value.string,x)==-1) return -1;
			*x = variable->value.floating;
			return 0;
	};

	return -1;
}

int VariableGetString(SVariable *variable, char *x) {
	switch(variable->type) {
		case VARIABLE_INTEGER:
			sprintf(x,"%d",variable->value.integer);
			return 0;
		case VARIABLE_FLOATING:
			sprintf(x,"%g",variable->value.floating);
			return 0;
		case VARIABLE_STRING:
			strcpy(x,variable->value.string);
			return 0;
	};

	return -1;
}

int VariableCreate(SVariable **variablelist, char* name, UVariableValue *value, int type) {
	SVariable *tmp1=NULL, *tmp2;

	tmp2 = *variablelist;
	while(tmp2 != NULL) {
		tmp1 = tmp2;
		tmp2 = tmp2->next;
	}
	/* tmp1 now points to the last element in the list */
	/* Now try to allocate memory for the new variable */
	if((tmp2 = (SVariable*) malloc(sizeof(SVariable))) == NULL) {
		/* Error, can't allocate memory. Return -1 */
		return -1;
	}

	strcpy(tmp2->name,name);
	tmp2->next = NULL;
	tmp2->type = type;
	if(tmp2->type==VARIABLE_STRING) {
		if((tmp2->value.string = (char*) malloc(MAX_STRING)) == NULL) {
			free(tmp2);
			return -1;
		}
		tmp2->value.string[0] = 0;
	} else {
		tmp2->value = *value;
	}

	if(tmp1 == NULL) {
		/* If the list was empty, let's make this the first item */
		*variablelist = tmp2;
	} else {
		/* If this isn't the first item, let the previous item point to this */
		tmp1->next = tmp2;
	}

	return 0;
}

SVariable* VariablePointer(SVariable *variablelist, char* name) {
	SVariable *tmp1;

	tmp1 = variablelist;
	while(tmp1 != NULL && stricmp(tmp1->name,name)) {
		/* Search through the list until the variable we're looking for, or
		   the end of list is found */
		tmp1 = tmp1->next;
	}
	if(tmp1 == NULL) {
		/* We found the last item, but not what we wanted, return NULL */
		return NULL;
	}
	return tmp1;
}

int VariableSet(SVariable *variablelist, char *name, UVariableValue *value) {
	SVariable *tmp1;

	if((tmp1 = VariablePointer(variablelist, name))	== NULL) {
		/* Can't find the variable, return -1 */
		return -1;
	}

	tmp1->value = *value;
	return 0;
}

int VariableGet(SVariable *variablelist, char *name, UVariableValue *value) {
	SVariable *tmp1;

	if((tmp1 = VariablePointer(variablelist, name))	== NULL) {
		/* Can't find the variable, return -1 */
		return -1;
	}

	*value = tmp1->value;
	return 0;
}

void VariableFree(SVariable *variablelist) {
	SVariable *tmp1, *tmp2;

	tmp1 = variablelist;
	while(tmp1 != NULL) {
		if(tmp1->type == VARIABLE_STRING) free(tmp1->value.string);
		tmp2 = tmp1;
		tmp1 = tmp1->next;
		free(tmp2);
	}
}
